/*
 * Copyright (c) 2004-2011 StreamBase Systems, Inc.  All rights reserved.
 * Author: Darach Ennis <darach.ennis@streambase.com>
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#include "SyncBsm.h"

#include <cutil_inline.h>
#include "BlackScholes_kernel.cu"

typedef struct _NativePointerMap_st {
  // Buffer sizes
  size_t optionSize;
  size_t resultSize;
  size_t s_option;
  size_t s_result;
  // CUDA context
  CUcontext cuCtx;
  // Result
  float *h_result;
  //CPU instance of input data
  float *h_CallResult;
  float *h_PutResult;
  float *h_StockPrice;
  float *h_OptionStrike;
  float *h_OptionYears;
  //GPU instance of input data
  float *d_StockPrice;
  float *d_OptionStrike;
  float *d_OptionYears;
  float *d_PutResult;
  float *d_CallResult;
} NativePointerMap;
    
/*
 * Class:     com_streambase_demos_nvidia_bsm_SyncBsm
 * Method:    initCuda
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_streambase_demos_nvidia_bsm_SyncBsm_initCuda
  (JNIEnv *env, jobject jObject)
{
    // Cache reference to calling java object instance
    //
    const jclass jClass = env->GetObjectClass(jObject); 

    // Cache reference to parameter fields
    //
    const jfieldID jOptionMicroBatchSize = env->GetFieldID(jClass, "optionMicroBatchSize", "I");
    const jfieldID jRiskFree = env->GetFieldID(jClass, "riskFree", "F");
    const jfieldID jVolatility = env->GetFieldID(jClass, "volatility", "F");
    const int optionSize = env->GetIntField(jObject,jOptionMicroBatchSize);
    const float riskFree = env->GetFloatField(jObject,jRiskFree);
    const float volatility = env->GetFloatField(jObject,jVolatility);

    // Compute native fields derived from parameter fields
    //
    const int resultSize = optionSize * 2;

    NativePointerMap* map = (NativePointerMap *)malloc(sizeof(NativePointerMap));

    // Calculate buffer sizes
    //
    map->optionSize = optionSize;
    map->resultSize = resultSize;
    map->s_result = resultSize * sizeof(float);
    map->s_option = optionSize * sizeof(float);

    // Allocate host buffers - once and only once, during init
    //
    map->h_CallResult = (float*)malloc(map->s_option);
    map->h_PutResult = (float*)malloc(map->s_option);
    map->h_StockPrice = (float*)malloc(map->s_option);
    map->h_OptionStrike = (float*)malloc(map->s_option);
    map->h_OptionYears = (float*)malloc(map->s_option);
    map->h_result = (float*)malloc(map->s_result);

    cuCtxCreate(&map->cuCtx, CU_CTX_SCHED_AUTO, cutGetMaxGflopsDeviceId() );

    // Allocate device buffers - once and only once, during init
    //
    cutilSafeCall( cudaMalloc((void **)&map->d_StockPrice,   map->s_option) );
    cutilSafeCall( cudaMalloc((void **)&map->d_OptionStrike, map->s_option) );
    cutilSafeCall( cudaMalloc((void **)&map->d_OptionYears, map->s_option) );
    cutilSafeCall( cudaMalloc((void **)&map->d_CallResult,   map->s_option) );
    cutilSafeCall( cudaMalloc((void **)&map->d_PutResult,    map->s_option) );

    // Cache buffer pointers
    //
    const jfieldID jNativePtrMap =
        env->GetFieldID(jClass, "nativePtrMap", "J");
    env->SetLongField(jObject, jNativePtrMap, (long)map);
}

/*
 * Class:     com_streambase_demos_nvidia_bsm_SyncBsm
 * Method:    callCuda
 * Signature: (Ljava/nio/ByteBuffer;)Ljava/nio/ByteBuffer;
 */
JNIEXPORT jobject JNICALL Java_com_streambase_demos_nvidia_bsm_SyncBsm_callCuda
  (JNIEnv *env, jobject jObject, jobject jParamsBuffer)
{
    // Cache reference to calling java object instance
    //     
    const jclass jClass = env->GetObjectClass(jObject);

    // Cache per-instance reference to parameter fields
    //
    const jfieldID jOptionMicroBatchSize = 
	env->GetFieldID(jClass, "optionMicroBatchSize", "I");
    const jfieldID jRiskFree = 
	env->GetFieldID(jClass, "riskFree", "F");
    const jfieldID jVolatility = 
	env->GetFieldID(jClass, "volatility", "F");
    const float riskFree = 
	env->GetFloatField(jObject,jRiskFree);
    const float volatility = 
	env->GetFloatField(jObject,jVolatility);

    // Cache per-instance reference to host memory
    //
    jfieldID jNativePtrMap = 
	env->GetFieldID(jClass, "nativePtrMap", "J");
    NativePointerMap* map = 
	(NativePointerMap *)env->GetLongField(jObject, jNativePtrMap);

    cuCtxSetCurrent(map->cuCtx);

    // Copy parameters
    //
    float* paramsBuf = (float*) env->GetDirectBufferAddress(jParamsBuffer);
    int paramsSize = env->GetDirectBufferCapacity(jParamsBuffer);
    for (int i = 0; i < map->optionSize; i++)
    {
        long j = i * 3;
        map->h_StockPrice[i] = paramsBuf[j];
        map->h_OptionStrike[i] = paramsBuf[j+1];
        map->h_OptionYears[i] = paramsBuf[j+2];
        map->h_CallResult[i] = 0.f; // reinit
        map->h_PutResult[i] = -1.f;  // reinit
    }

    // Copy input data to GPU memory
    //
    cutilSafeCall( cudaMemcpy(map->d_CallResult, map->h_CallResult, map->s_option, cudaMemcpyHostToDevice) );
    cutilSafeCall( cudaMemcpy(map->d_PutResult, map->h_PutResult, map->s_option, cudaMemcpyHostToDevice) );
    cutilSafeCall( cudaMemcpy(map->d_StockPrice,  map->h_StockPrice,   map->s_option, cudaMemcpyHostToDevice) );
    cutilSafeCall( cudaMemcpy(map->d_OptionStrike, map->h_OptionStrike,  map->s_option, cudaMemcpyHostToDevice) );
    cutilSafeCall( cudaMemcpy(map->d_OptionYears,  map->h_OptionYears,   map->s_option, cudaMemcpyHostToDevice) );

    // Call GPU kernel
    //
    BlackScholesGPU<<<1024, 64>>>(
        map->d_CallResult,
        map->d_PutResult,
        map->d_StockPrice,
        map->d_OptionStrike,
        map->d_OptionYears,
        riskFree,
        volatility,
        map->optionSize
    );
    cutilCheckMsg("BlackScholesGPU() execution failed\n");

    float * resultIdx = (float *) (map->h_result + map->optionSize );

    cutilSafeCall( cudaMemcpy(map->h_CallResult, map->d_CallResult, map->s_option, cudaMemcpyDeviceToHost ) );
    cutilSafeCall( cudaMemcpy(map->h_PutResult,  map->d_PutResult,  map->s_option, cudaMemcpyDeviceToHost ) );

    memcpy(map->h_result, map->h_PutResult, map->s_option);
    memcpy(resultIdx, map->h_CallResult, map->s_option);
      
    return env->NewDirectByteBuffer(map->h_result, map->s_result);
}

/*
 * Class:     com_streambase_demos_nvidia_bsm_SyncBsm
 * Method:    freeCuda
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_streambase_demos_nvidia_bsm_SyncBsm_freeCuda
  (JNIEnv *env, jobject jObject)
{
    jclass jClass = env->GetObjectClass(jObject);

    // Cache per-instance reference to cached buffers
    jfieldID jNativePtrMap = env->GetFieldID(jClass, "nativePtrMap", "J");
    NativePointerMap* map = (NativePointerMap *)env->GetLongField(jObject, jNativePtrMap);

    cutilSafeCall( cudaFree(map->d_StockPrice) );
    cutilSafeCall( cudaFree(map->d_OptionStrike) );
    cutilSafeCall( cudaFree(map->d_OptionYears) );
    cutilSafeCall( cudaFree(map->d_PutResult ) );
    cutilSafeCall( cudaFree(map->d_CallResult ) );

    free(map->h_OptionYears);
    free(map->h_OptionStrike);
    free(map->h_StockPrice);
    free(map->h_PutResult);
    free(map->h_CallResult);
    free(map->h_result);

    cuCtxDestroy(map->cuCtx);

    free(map);
}
